/////////////////////////////////////////////////////////////////////////////////////
// WORKER THREAD DERIVED CLASS - produced by Worker Thread Class Generator Rel. 1.0
// ThreadSearch.cpp: Implementation of the CThreadSearch Class.
/////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ThreadSearch.h"
#include <afxtempl.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

IMPLEMENT_DYNAMIC(CThreadSearch, CThread)

/////////////////////////////////////////////////////////////////////////////////////
// Construction/Destruction
/////////////////////////////////////////////////////////////////////////////////////
CThreadSearch::CThreadSearch(CListCtrl *result, CListBox *status,
							 SEARCH_STATE *search_state,
							 CListCtrlHelper *helper,
							 DATABASE * database,
							 BOOL ping_hosts,
							 CPing *ping,
							 void* pOwnerObject, LPARAM lParam)
	: CThread(pOwnerObject, lParam),
	  m_lcResult(result), m_lbStatus(status), 
	  m_sqlresult(NULL),
	  m_psearch_state(search_state),
	  m_phelp(helper),
	  m_database(database),
	  m_ping_hosts(ping_hosts),
	  m_ping(ping)
{
	// WORKER THREAD CLASS GENERATOR - Support Thread Notification //////////////////
	SUPPORT_THREAD_NOTIFICATION
	// User Initialization here...
}

CThreadSearch::~CThreadSearch()
{}

/////////////////////////////////////////////////////////////////////////////////////
// CThreadSearch diagnostics
#ifdef _DEBUG
void CThreadSearch::AssertValid() const
{
	CThread::AssertValid();
}

void CThreadSearch::Dump(CDumpContext& dc) const
{
	CThread::Dump(dc);
}
#endif //_DEBUG
/////////////////////////////////////////////////////////////////////////////////////

// Unallocate all thread specific extra resources if needed while killing this thread
void CThreadSearch::OnKill()
{
	// Code here...
}

/////////////////////////////////////////////////////////////////////////////////////
// WORKER THREAD CLASS GENERATOR - Do not remove this method!
// MAIN THREAD HANDLER - The only method that must be implemented.
/////////////////////////////////////////////////////////////////////////////////////
DWORD CThreadSearch::ThreadHandler()
{
	BOOL bContinue = TRUE;
	int nIncomingCommand;

	do
	{
		WaitForNotification(nIncomingCommand/*, CThreadSearch::DEFAULT_TIMEOUT*/);

		/////////////////////////////////////////////////////////////////////////////
		//	Main Incoming Command Handling:
		/////////////////////////////////////////////////////////////////////////////
		switch (nIncomingCommand)
		{

		case CThread::CMD_TIMEOUT_ELAPSED:
			if (GetActivityStatus() != CThread::THREAD_PAUSED)
			{
				SetActivityStatus(CThread::THREAD_PENDING);
				// UserSpecificTimeoutElapsedHandler();
				// Run();					// fire CThread::CMD_RUN Command if needed
			};
			break;

		case CThread::CMD_INITIALIZE:		// initialize Thread Task
			// this Command must be handled; it is automatically fired when the Thread starts
			SetActivityStatus(CThread::THREAD_RUNNING);
			// UserSpecificOnInitializeHandler();
			Run();							// fire CThread::CMD_RUN Command after proper initialization
			break;

		case CThread::CMD_RUN:				// handle 'OnRun' Command
			SetActivityStatus(CThread::THREAD_RUNNING);

			Search();

			SetActivityStatus(CThread::THREAD_PENDING);
			break;

		case CThread::CMD_PAUSE:			// handle 'OnPause' Command
			if (GetActivityStatus() == CThread::THREAD_RUNNING)
			{
				SetActivityStatus(CThread::THREAD_PAUSED);
				// UserSpecificOnPauseHandler();
			};
			break;

		case CThread::CMD_CONTINUE:			// handle 'OnContinue' Command
			if (GetActivityStatus() == CThread::THREAD_PAUSED)
			{
				SetActivityStatus(CThread::THREAD_CONTINUING);
				// UserSpecificOnContinueHandler();
				// Run();					// fire CThread::CMD_RUN Command if needed
			};
			SetActivityStatus(CThread::THREAD_PENDING);
			break;

		/*
		case CThreadSearch::CMD_USER_SPECIFIC:	// handle the user-specific Command
			SetActivityStatus(CThreadSearch::THREAD_USER_SPECIFIC);
			UserSpecificOnUserCommandHandler();
			SetActivityStatus(CThread::THREAD_PENDING);
			break;
		*/

		case CThread::CMD_STOP:				// handle 'OnStop' Command
			if ( m_sqlresult != NULL)
				mysql_free_result( m_sqlresult );
			// UserSpecificOnStopHandler();
			bContinue = FALSE;				// ... and leave thread function finally
			break;

		default:							// handle other Commands...
			break;

		};

	} while (bContinue);

	return (DWORD)CThread::DW_OK;	//... if Thread task completition OK
}

void CThreadSearch::Search()
{
	unsigned int results;
	bool more_available;
	CString limit, tmp, minstr, addstr;
	CTime t1 = CTime::GetCurrentTime();
	CStdioFile file;
	CFileStatus status;

	m_strQuery = m_psearch_state->search;
	m_substr = m_psearch_state->subsearch;
	m_lcResult->DeleteAllItems();
	m_psearch_state->first_old = m_psearch_state->first;
	m_psearch_state->submit->SetWindowText("Stop");

	// connect to database
	CMySql mysql(m_database->host, m_database->login, 
				 m_database->password, m_database->dbname);

	if (!m_psearch_state->search_dir)
	{
		results = SearchFiles(mysql, more_available,
							  m_psearch_state->first, m_psearch_state->hits_per_page);
		if (m_psearch_state->quit)
			return;
		m_psearch_state->first += results;
		if (results < m_psearch_state->hits_per_page)
		{
			m_psearch_state->search_dir = true;
			m_psearch_state->first = SearchPath(mysql, more_available, 0,
												m_psearch_state->hits_per_page-results);
			results += m_psearch_state->first;
		}
		else if (more_available == false)
			// we got enough files but no more available
			// before we jump to any conclusions, check if there are an folders
			// available!
			SearchPath(mysql, more_available, 0, 0);
			if (m_psearch_state->quit)
				return;

	}
	else
	{
		if (m_psearch_state->quit)
			return;
		results = SearchPath(mysql, more_available,
							 m_psearch_state->first, m_psearch_state->hits_per_page);
		m_psearch_state->first += results;
	}

	// just for safety - if more available then result buffer should be full too
	if (more_available)
		ASSERT (results == m_psearch_state->hits_per_page);

	m_psearch_state->more_available = more_available;
	m_psearch_state->num_page++;

	// measure time
	CTime t2 = CTime::GetCurrentTime();
	CTimeSpan time_needed = t2 - t1;

	minstr.Format("%u min ", time_needed.GetMinutes());

	if (results > 0)
		addstr.Format("Treffer %u - %u", m_psearch_state->first_old,
										 m_psearch_state->first_old + results - 1);
	else
		addstr = "Keine Treffer";
	tmp.Format("[%s] %s (%s%u sec)", m_psearch_state->str_query,
									 addstr,
								     time_needed.GetMinutes() > 0 ? minstr : "",
									 time_needed.GetSeconds());
	m_lbStatus->SetCurSel(m_lbStatus->AddString(tmp));

	tmp.Format("%06d", m_psearch_state->num_page);

	if (file.Open(m_psearch_state->path + tmp, CFile::modeCreate | CFile::modeWrite | CFile::shareDenyNone))
	{
		m_phelp->ListControlToFile(file, m_psearch_state->first_old);
		file.GetStatus(status);
		status.m_attribute = 0x06; // make the file system and hidden
		file.Close();
		m_psearch_state->max_page_cached = m_psearch_state->num_page;
		try
		{
			CFile::SetStatus(m_psearch_state->path + tmp, status);
		}
		catch (CFileException *e) 
		{
			e->Delete(); 
		}
	}
	else
	{
		AfxMessageBox(IDS_ERR_CACHEFILE_CREATE);
		PostQuitMessage(-1);
	}
	m_psearch_state->forward->EnableWindow(more_available ? TRUE : FALSE);
	m_psearch_state->rewind->EnableWindow(m_psearch_state->num_page > 0);
	if (more_available)
		m_psearch_state->forward->SetFocus();
	else
		m_psearch_state->submit->SetFocus();

	m_psearch_state->submit->SetWindowText("Suchen");
}

unsigned int CThreadSearch::SearchFiles(CMySql &mysql, bool &more_available,
									   unsigned int first, unsigned int num)
{
	BOOL resultValid, inserted, first_pid = TRUE;
	unsigned int pid, i;

	CString pidstring, tmp, type, host, error, limit;
	CMap <int, int, BOOL, BOOL> pids_registered;
	CMap <int, int, PATHINFO, PATHINFO> path_cache;
	CArray <FILEINFO, FILEINFO> file_cache;

	MYSQL_ROW row;
	PATHINFO path_data;
	FILEINFO fileinfo;

	limit.Format(" LIMIT %u, %u", first, num + 1);
	more_available = false;

	tmp.LoadString(IDS_SQL_SELECT_FILE);

	if (m_psearch_state->quit)
		return 0;

	if (!mysql.Query(tmp + ' ' + m_strQuery + m_substr + ' ' + limit))
	{
		mysql.LastError(error);
		tmp.LoadString(IDS_ERR_SQL_QUERY);
		AfxMessageBox(tmp + error);
		return 0;
	}

	if (m_psearch_state->quit)
		return 0;

	// retrieve result
	i = 0;
	while ( i < num && mysql.GetNextRow(row, resultValid) && resultValid )
	{
		for (int j = 0; j < 4; j++)
			if (row[j] == NULL)
				return 0;

		// save data to file_cache for later use
		pid = atoi(row[0]);
		fileinfo.PID = pid;
		fileinfo.FileName = row[1];
		fileinfo.FileSize = atoi(row[2]);
		fileinfo.FileDate = row[3];
		file_cache.Add(fileinfo);

		// check if we have this PID already
		if (pids_registered.Lookup(pid, inserted) == 0)
		{
			// element not yet included, so add it to the string & map
			pids_registered.SetAt(pid, TRUE);
			tmp.Format("PID=%d", pid);
			if (!first_pid)
				pidstring += " OR ";
			pidstring += tmp;
		}
		first_pid = FALSE;
		i++;
	}

	// fetch another row to see if there is more for us 
	more_available = (mysql.GetNextRow(row, resultValid) && resultValid && i == num);

	// return if nothing found
	if (i == 0)
		return 0;

	// get all pathinfos
	tmp.LoadString(IDS_SQL_SELECT_PID);
	if (!mysql.Query(tmp + ' ' + pidstring))
	{
		mysql.LastError(error);
		tmp.LoadString(IDS_ERR_SQL_QUERY);
		AfxMessageBox(tmp + error);
		return 0;
	}

	i = 0;

	if (m_psearch_state->quit)
		return 0;

	/*
SELECT Host.HostName, ShareName, PathName, Host.HostType, Path.PID,
 Login, PassWord, Port from Path natural left join Share natural left join Host natural left join FTP  WHERE	
	*/
	while ( mysql.GetNextRow(row, resultValid) && resultValid )
	{
		for (int j = 0; j < 5; j++)
			if (row[j] == NULL)
				return 0;

		pid = atoi(row[4]);
		path_data.HostName = row[0];
		path_data.ShareName = row[1];
		path_data.PathName = row[2];
		path_data.HostType = atoi(row[3]);
		if (atoi(row[3]) == HOST_FTP)
		{
			path_data.Login = (row[5] == NULL) ? "" : row[5];
			path_data.PassWord = (row[6] == NULL) ? "" : row[6];
			path_data.Port = (row[7] == NULL) ? "" : row[7];
		}

		// ... and put it in the map
		path_cache.SetAt(pid, path_data);
		i++;
	}
	
	if (i == 0)
	{
		tmp.LoadString(IDS_ERR_FATAL);
		AfxMessageBox(tmp + " [Pathinfo empty]");
		return 0;
	}

	i = 0;
	// iterate through the cache, get pathinfo and add everything to box
	for (int j = 0; j < file_cache.GetSize(); j++)
	{
		fileinfo = file_cache[j];

		// if the pid is not in the cache something went seriously wrong, better quit
		if (path_cache.Lookup(fileinfo.PID, path_data) == 0)
		{
			tmp.LoadString(IDS_ERR_FATAL);
			AfxMessageBox(tmp + " [Path Cache Miss]");
			return 0;
		}

		// path is displayed different depending on the type
		if (path_data.HostType == HOST_SMB)
		{
			tmp = path_data.ShareName +
				  path_data.PathName + '/' + fileinfo.FileName;
			type = "SMB";
			host = path_data.HostName;
		}
		else
		{
			tmp = path_data.PathName + '/' + fileinfo.FileName;
			host = path_data.Login + ':' + path_data.PassWord + '@' +
				   path_data.HostName + ':' + path_data.Port;
			type = "FTP";
		}

		int icon = -1;
		if (m_ping_hosts)
		{
			if (IsOnline(path_data.HostName))
				icon = m_phelp->GetIndexIcon(tmp);
		}
		else
			icon = m_phelp->GetIndexIcon(tmp);

		m_phelp->AddRow(i, tmp, icon, m_phelp->GetSmartFileSize(fileinfo.FileSize), 
						host, type);
		i++;
	}

	return i;
}

unsigned int CThreadSearch::SearchPath(CMySql &mysql, bool &more_available,
									   unsigned int first, unsigned int num)
{
	CString error, tmp, host, type, limit;
	MYSQL_ROW row;
	BOOL resultValid;
	unsigned int i = 0;

	MakePathQuery(m_strQuery);
	limit.Format(" LIMIT %u, %u", first, num + 1);
	tmp.LoadString(IDS_SQL_SELECT_PATH);
	if (!mysql.Query(tmp + ' ' + m_strQuery + limit))
	{
		mysql.LastError(error);
		tmp.LoadString(IDS_ERR_SQL_QUERY);
		AfxMessageBox(tmp + error);
		more_available = false;
		return 0;
	}

	if (m_psearch_state->quit)
		return 0;

	// fetch num - 1 rows
	while ( i < num && mysql.GetNextRow(row, resultValid) && resultValid )
	{
		if (atoi(row[3]) == HOST_SMB)
		{
			tmp = row[1];
			tmp += row[2];
			type = "SMB";
			host = row[0];
		}
		else
		{
			tmp = row[2];
			host = row[4];
			host +=  ':';
			host += row[5];
			host +=  '@';
			host += row[0];
			host += ':';
			host += row[6];
			type ="FTP";
		}

		int icon = ICON_FOLDER;
		if (m_ping_hosts && !IsOnline(row[0]))
			icon = -1;

		m_phelp->AddRow(i, tmp, icon, "", host, type);
		i++;
	}
	
	// fetch another row to see if there is more for us 
	more_available = (mysql.GetNextRow(row, resultValid) && resultValid && i == num);
		
	return i;
}

void CThreadSearch::MakePathQuery(CString &query)
{
	CString result;

	query.Replace("FileName", "Host.HostName");
	result = query + " OR ";

	query.Replace("Host.HostName", "ShareName");
	result += query + " OR ";

	query.Replace("ShareName", "PathName");
	result += query;

	query = '(' + result + ')';
}

BOOL CThreadSearch::IsOnline(CString host)
{
	BOOL found = FALSE, host_online = FALSE;
	int i = 0;

	while (!found && i < online.GetSize())
		if (online[i].host == host)
			found = TRUE;
		else
			i++;

	if (found)
		return online[i].online;
	else
	{
		ONLINE_ENTRY entry;

		entry.host = host;
		
		entry.online = m_ping->Ping(host);
		online.Add(entry);
		return entry.online;
	}
}
